//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <list>
#include <string>
#include <sstream>
#include "PluginInterface.h"
#include "Scintilla.h"

#pragma argsused

typedef std::list<std::string> TStringList;

// This is the name which will be displayed in Plugins Menu
const char PLUGIN_NAME[] = "Sort";

NppData nppData;

// The number of functionalities
const int nbFunc = 6;
// Declaration of functionality (FuncItem) Array
FuncItem funcItem[nbFunc];


// Here are the declaration of functions for 3 functionalities
void __cdecl SortLinesInSelection();
void __cdecl ReverseSortLinesInSelection();
void __cdecl ToggleCaseSensitivity();
void __cdecl KeyOnEntireLine();
void __cdecl KeyOnCursorColumn();

void __cdecl DoNothing(); // Dummy routine for menu separator targets

bool CaseSensitiveComparison = false;
bool CompareFromCursorColumn = false;
unsigned int ComparisonColumn = 0;

int APIENTRY DllEntryPoint(HINSTANCE hinst, unsigned long reason, void* lpReserved)
{
  switch (reason)
  {
    // Here we initialize the FuncItem Array.
    // AWARE : no need to initialize _cmdID field.
    // This filed will be initialized by Notepad++ plugins system
    case DLL_PROCESS_ATTACH:
      funcItem[0]._pFunc = SortLinesInSelection;
      funcItem[1]._pFunc = ReverseSortLinesInSelection;
      funcItem[2]._pFunc = ToggleCaseSensitivity;
      funcItem[3]._pFunc = DoNothing;
      funcItem[4]._pFunc = KeyOnEntireLine;
      funcItem[5]._pFunc = KeyOnCursorColumn;

      strcpy(funcItem[0]._itemName, "Ascending");
      strcpy(funcItem[1]._itemName, "Descending");
      strcpy(funcItem[2]._itemName, "Use Case-sensitivity");
      strcpy(funcItem[3]._itemName, "-");
      strcpy(funcItem[4]._itemName, "Compare entire line");
      strcpy(funcItem[5]._itemName, "Compare from selection column");

      // if you want your menu item is checked while the launch of program
      // set it as true, otherwise set it as false
      funcItem[0]._init2Check = false;
      funcItem[1]._init2Check = false;
      funcItem[2]._init2Check = CaseSensitiveComparison;
      funcItem[3]._init2Check = false;
      funcItem[4]._init2Check = !CompareFromCursorColumn;
      funcItem[5]._init2Check = CompareFromCursorColumn;
      break;

    case DLL_PROCESS_DETACH:
      break;

    case DLL_THREAD_ATTACH:
      break;

    case DLL_THREAD_DETACH:
      break;
  }

  return true;
}
//---------------------------------------------------------------------------


// The setInfo function gets the needed infos from Notepad++ plugins system
void DLLIMPORT __cdecl setInfo(struct NppData notpadPlusData) {

  nppData = notpadPlusData;
}

// The getName function tells Notepad++ plugins system its name
const char DLLIMPORT * __cdecl getName(void) {

  return PLUGIN_NAME;
}

// The getFuncsArray function gives Notepad++ plugins system the pointer FuncItem Array
// and the size of this array (the number of functions)
struct FuncItem DLLIMPORT * __cdecl getFuncsArray(int *nbF) {

  *nbF = nbFunc;
  return funcItem;
}

// If you don't need notifications from Notepad++,
// just let it be empty.
void DLLIMPORT __cdecl beNotified(struct SCNotification *notifyCode) {
  static unsigned runonce=0;
  if (!runonce) {
    MENUITEMINFO miosep={sizeof(MENUITEMINFO), MIIM_TYPE,MFT_SEPARATOR,0,0,0,0,0,0,NULL,0};
    for(runonce=0; runonce<nbFunc; runonce++)
      if (funcItem[runonce]._itemName[0]=='-')
        ::SetMenuItemInfo(GetMenu(nppData._nppHandle), funcItem[runonce]._cmdID, FALSE, &miosep);
    DrawMenuBar(nppData._nppHandle);
  }
}

//----------------------------------------------------------//
// Here're the definition of functions for 3 functionalities
//----------------------------------------------------------//

bool CaseSensitiveLess(const std::string& a, const std::string& b)
{
  return strcmp(a.c_str() + std::min(ComparisonColumn, a.length()),
                b.c_str() + std::min(ComparisonColumn, b.length())) < 0;
}
//---------------------------------------------------------------------------


bool CaseInsensitiveLess(const std::string& a, const std::string& b)
{
  return stricmp(a.c_str() + std::min(ComparisonColumn, a.length()),
                 b.c_str() + std::min(ComparisonColumn, b.length())) < 0;
}
//---------------------------------------------------------------------------


void SortLinesInSelection(bool ascending)
{
  int currentEdit;
  ::SendMessage(nppData._nppHandle, WM_GETCURRENTSCINTILLA, 0, reinterpret_cast<LPARAM>(&currentEdit));
  HWND TgtWindow = nppData._scintillaMainHandle;
  if (currentEdit)
    TgtWindow = nppData._scintillaSecondHandle;

  // Find the current selection, if any
  TextRange SortRange;
  LRESULT StartSelection = ::SendMessage(TgtWindow, SCI_GETSELECTIONSTART, 0, 0);
  LRESULT EndSelection = ::SendMessage(TgtWindow, SCI_GETSELECTIONEND, 0, 0);

  if (EndSelection == StartSelection) { // No selection -- sort the entire buffer
    SortRange.chrg.cpMin = 0;
    SortRange.chrg.cpMax = ::SendMessage(TgtWindow, SCI_GETLENGTH, 0, 0);
  }
  else { // Find the start of the first and end of the last lines in the selection
    SortRange.chrg.cpMin = ::SendMessage(TgtWindow, SCI_POSITIONFROMLINE,
        ::SendMessage(TgtWindow, SCI_LINEFROMPOSITION, StartSelection, 0), 0);
    SortRange.chrg.cpMax = ::SendMessage(TgtWindow, SCI_POSITIONFROMLINE,
        ::SendMessage(TgtWindow, SCI_LINEFROMPOSITION, EndSelection, 0)+1, 0);
  }

  // Get the text from those lines
  std::string SortText(SortRange.chrg.cpMax - SortRange.chrg.cpMin + 1, ' ');
  SortRange.lpstrText = const_cast<char *>(SortText.data());
  ::SendMessage(TgtWindow, SCI_GETTEXTRANGE, 0, reinterpret_cast<LPARAM>(&SortRange));

  // Parse the text into the individual lines
  std::string::size_type cr = SortText.find('\r');
  std::string::size_type lf = SortText.find('\n');
  char LastTerminator = '\n';
  if (lf == std::string::npos)
    LastTerminator = '\r';
  TStringList Lines;
  std::istringstream is(SortText.c_str());
  is.clear();
  while (is.good() && !is.str().empty()) {
    std::string ThisLine;
    std::getline(is, ThisLine, LastTerminator);
    if (!ThisLine.empty()) {
      if ((ThisLine[ThisLine.length()-1] == '\n') || (ThisLine[ThisLine.length()-1] == '\r'))
        ThisLine.erase(ThisLine.length()-1,1);
      Lines.push_back(ThisLine);
    }
  }

  // Choose the character-column to start the comparison
  ComparisonColumn = 0;
  if (CompareFromCursorColumn)
    ComparisonColumn = ::SendMessage(TgtWindow, SCI_GETCOLUMN, StartSelection, 0);

  // Sort the lines
  if (CaseSensitiveComparison)
    Lines.sort(CaseSensitiveLess);
  else
    Lines.sort(CaseInsensitiveLess);
  if (!ascending)
    Lines.reverse();

  // Decide what line terminator to use in reassembling the sorted lines
  char LineTerminator[3];
  int i = 0;
  if (cr != std::string::npos)
    LineTerminator[i++] = '\r';
  if (lf != std::string::npos)
    LineTerminator[i++] = '\n';
  LineTerminator[i] = '\0';

  // Reassemble the sorted lines into a single string
  std::ostringstream os;
  for (TStringList::iterator line = Lines.begin(); line != Lines.end(); line++)
    os << *line << LineTerminator;

  // Mark as a single undo operation
  ::SendMessage(TgtWindow, SCI_BEGINUNDOACTION, 0, 0);

  // Mark the unsorted lines as a "target" so we can replace the target with the sorted lines.
  ::SendMessage(TgtWindow, SCI_SETTARGETSTART, SortRange.chrg.cpMin, 0);
  ::SendMessage(TgtWindow, SCI_SETTARGETEND, SortRange.chrg.cpMax, 0);

  ::SendMessage(TgtWindow, SCI_REPLACETARGET, os.str().length(), reinterpret_cast<LPARAM>(os.str().c_str()));

  ::SendMessage(TgtWindow, SCI_ENDUNDOACTION, 0, 0);
}
//---------------------------------------------------------------------------


void __cdecl SortLinesInSelection()
{
  SortLinesInSelection(true);
}
//---------------------------------------------------------------------------


void __cdecl ReverseSortLinesInSelection()
{
  SortLinesInSelection(false);
}
//---------------------------------------------------------------------------


void __cdecl ToggleCaseSensitivity()
{
  CaseSensitiveComparison = !CaseSensitiveComparison;
  ::CheckMenuItem(::GetMenu(nppData._nppHandle), funcItem[2]._cmdID,
      MF_BYCOMMAND | (CaseSensitiveComparison ? MF_CHECKED : MF_UNCHECKED));
}
//---------------------------------------------------------------------------


void __cdecl UpdateColumnMenuItems()
{
  ::CheckMenuItem(::GetMenu(nppData._nppHandle), funcItem[4]._cmdID,
      MF_BYCOMMAND | (CompareFromCursorColumn ? MF_UNCHECKED : MF_CHECKED));
  ::CheckMenuItem(::GetMenu(nppData._nppHandle), funcItem[5]._cmdID,
      MF_BYCOMMAND | (CompareFromCursorColumn ? MF_CHECKED : MF_UNCHECKED));
}
//---------------------------------------------------------------------------



void __cdecl KeyOnEntireLine()
{
  CompareFromCursorColumn = false;
  UpdateColumnMenuItems();
}
//---------------------------------------------------------------------------


void __cdecl KeyOnCursorColumn()
{
  CompareFromCursorColumn = true;
  UpdateColumnMenuItems();
}
//---------------------------------------------------------------------------


void __cdecl DoNothing()
{
} // Dummy routine for menu separator targets

